# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"map_dict_ex": "00_core.ipynb",
         "D": "00_core.ipynb",
         "tensor2shape": "00_core.ipynb",
         "tensor2mu": "00_core.ipynb",
         "tensor2mu.__docs__": "00_core.ipynb",
         "TensorBatch": "00_core.ipynb",
         "obj2tensor": "00_core.ipynb",
         "BD": "00_core.ipynb",
         "dependencies": "00_nbdev_extension.ipynb",
         "create_conda_yaml": "00_nbdev_extension.ipynb",
         "create_conda_yamls": "00_nbdev_extension.ipynb",
         "Loop": "03_callback.core.ipynb",
         "defaults.loop_callbacks": "03_callback.core.ipynb",
         "LoopCallback": "03_callback.core.ipynb",
         "Agent": "04_agent.ipynb",
         "AgentCallback": "04_agent.ipynb",
         "TstCallback": "05c_data.gym.ipynb",
         "Src": "05a_data.block.ipynb",
         "SrcCallback": "05a_data.block.ipynb",
         "GymSrc": "05a_data.block.ipynb",
         "SrcProfiler": "05a_data.block.ipynb",
         "FirstLast": "05c_data.gym.ipynb",
         "SourceDataset": "05b_data.block_simple.ipynb",
         "IterableTfmdLists": "05b_data.block_simple.ipynb",
         "vstack_collate": "05b_data.block_simple.ipynb",
         "fr_collate": "05b_data.block_simple.ipynb",
         "IterableTfmdDL": "05b_data.block_simple.ipynb",
         "IterableDataBlock": "05b_data.block_simple.ipynb",
         "Datasets.__iter__": "05a_data.block.ipynb",
         "Datasets.__init__": "05b_data.block_simple.ipynb",
         "TransformBlock": "05b_data.block_simple.ipynb",
         "SourceBlock": "05a_data.block.ipynb",
         "map_iters": "05a_data.block.ipynb",
         "noop_splitter": "05a_data.block.ipynb",
         "get_all_keys": "05a_data.block.ipynb",
         "SourceDataBlock": "05b_data.block_simple.ipynb",
         "parse_events": "05b_data.block_simple.ipynb",
         "Source": "05c_data.gym.ipynb",
         "VerboseChecked": "05b_data.block_simple.ipynb",
         "IterableDatasets": "05b_data.block_simple.ipynb",
         "get_sources": "05b_data.block_simple.ipynb",
         "Callback.after_create": "05b_data.block_simple.ipynb",
         "Recorder.after_batch": "05b_data.block_simple.ipynb",
         "return_data": "05c_data.gym.ipynb",
         "ReturnHistory": "05c_data.gym.ipynb",
         "np_type_cast": "05c_data.gym.ipynb",
         "GymLoop": "05c_data.gym.ipynb",
         "FirstLastTfm": "05c_data.gym.ipynb",
         "Reward": "05c_data.gym.ipynb",
         "NEpisodes": "05c_data.gym.ipynb",
         "DQN": "10a_agents.dqn.core.ipynb",
         "ArgMaxFeed": "10a_agents.dqn.core.ipynb",
         "DiscreteEpsilonRandomSelect": "10a_agents.dqn.core.ipynb",
         "Epsilon": "10a_agents.dqn.core.ipynb",
         "ExperienceReplay": "10a_agents.dqn.core.ipynb",
         "DQNTrainer": "10a_agents.dqn.core.ipynb",
         "DQNTargetTrainer": "10b_agents.dqn.targets.ipynb",
         "DoubleDQNTrainer": "10c_agents.dqn.double.ipynb",
         "DuelingBlock": "10d_agents.dqn.dueling.ipynb",
         "DuelingDQN": "10d_agents.dqn.dueling.ipynb",
         "create_support": "10e_agents.dqn.categorical.ipynb",
         "CategoricalDQN": "10e_agents.dqn.categorical.ipynb",
         "distribute": "10e_agents.dqn.categorical.ipynb",
         "final_distribute": "10e_agents.dqn.categorical.ipynb",
         "categorical_update": "10e_agents.dqn.categorical.ipynb",
         "show_q_distribution": "10e_agents.dqn.categorical.ipynb",
         "PartialCrossEntropy": "10e_agents.dqn.categorical.ipynb",
         "CategoricalDQNTrainer": "10e_agents.dqn.categorical.ipynb",
         "CategoricalArgMaxFeed": "10e_agents.dqn.categorical.ipynb",
         "show_q": "10e_agents.dqn.categorical.ipynb",
         "show_q_and_max_distribution": "10e_agents.dqn.categorical.ipynb",
         "Critic": "11a_agents.policy_gradient.ppo.ipynb",
         "ClampBlock": "11a_agents.policy_gradient.ppo.ipynb",
         "show_multisurface": "11a_agents.policy_gradient.ppo.ipynb",
         "ConstBlock": "11a_agents.policy_gradient.ppo.ipynb",
         "clamp_or_const": "11a_agents.policy_gradient.ppo.ipynb",
         "show_surface": "11a_agents.policy_gradient.ppo.ipynb",
         "SimpleGMM": "11a_agents.policy_gradient.ppo.ipynb",
         "MultiCompGMM": "11a_agents.policy_gradient.ppo.ipynb",
         "show_install": "20_test_utils.ipynb"}

modules = ["core.py",
           "nbdev_extensions.py",
           "callback/core.py",
           "agent.py",
           "data/block.py",
           "data/block_simple.py",
           "data/gym.py",
           "agents/dqn/core.py",
           "agents/dqn/targets.py",
           "agents/dqn/double.py",
           "agents/dqn/dueling.py",
           "agents/dqn/categorical.py",
           "agents/policy_gradient/ppo.py",
           "test_utils.py"]

doc_url = "https://josiahls.github.io/fastrl/"

git_url = "https://github.com/josiahls/fastrl/tree/master/"

def custom_doc_links(name): return None
